from dasy import parser
from vyper.ast.nodes import If, Assign, Return
from .utils import next_nodeid

def parse_if(expr):
    return If(ast_type='If', node_id=next_nodeid(), test=parser.parse_node(expr[1]), body=[parser.parse_node(expr[2])], orelse=[parser.parse_node(expr[3])])

def parse_assignment(expr):
    match expr[1:]:
        case [target, value]:
            return Assign(ast_type='Call', node_id=next_nodeid(), targets=[parser.parse_node(target)], value=parser.parse_node(value))

def parse_return(return_tree):
    val = return_tree[1]
    value_node = parser.parse_node(val)
    return_node = Return(value=value_node, ast_type='Return', node_id=next_nodeid())
    return return_node



from hy import models
from dasy import parser
from .utils import next_nodeid
import vyper.ast.nodes as vy_nodes

COMP_FUNCS = ['<', '<=', '>', '>=', '==', '!=']

def chain_comps(expr):
    new_node = models.Expression()
    new_expr = [models.Symbol("and")]
    for vals in zip(expr[1:], expr[2:]):
        new_expr.append(models.Expression((expr[0], vals[0], vals[1])))
    new_node += tuple(new_expr)
    return new_node


def parse_comparison(comp_tree):
    if len(comp_tree[1:]) > 2: # comparing more than 2 things; chain comps for (< 2 3 4 )
        return parser.parse_node(chain_comps(comp_tree))
    left = parser.parse_node(comp_tree[1])
    right = parser.parse_node(comp_tree[2])
    op = parser.parse_node(comp_tree[0])
    return vy_nodes.Compare(left=left, ops=[op], comparators=[right], node_id=next_nodeid(), ast_type='Compare')



import ast as py_ast

import hy
import vyper.ast.nodes as vy_nodes
from hy import models

from .builtins import parse_builtin
from .core import (parse_attribute, parse_call, parse_contract,
                   parse_declaration, parse_declarations, parse_fn, parse_tuple)
from .ops import (BIN_FUNCS, BOOL_OPS, COMP_FUNCS, UNARY_OPS, parse_binop,
                  parse_boolop, parse_comparison, parse_unary)
from .stmt import parse_assignment, parse_if, parse_return
from .utils import next_nodeid

BUILTIN_FUNCS = BIN_FUNCS + COMP_FUNCS + UNARY_OPS + BOOL_OPS

NAME_CONSTS = ["True", "False"]

def parse_expr(expr):

    match expr:
        case models.Keyword(name), models.Integer(length):
            if str(name) == "string":
                value_node = parse_node(models.Symbol("String"))
            elif str(name) == "bytes":
                value_node = parse_node(models.Symbol("Bytes"))
            annotation = vy_nodes.Subscript(ast_type='Subscript', node_id=next_nodeid(), slice=vy_nodes.Index(ast_type='Index', node_id=next_nodeid(), value=parse_node(length)), value=value_node)
            annotation._children.add(value_node)
            return annotation

    
    cmd_str = str(expr[0])

    if cmd_str in BIN_FUNCS:
        return parse_binop(expr)
    if cmd_str in COMP_FUNCS:
        return parse_comparison(expr)
    if cmd_str in UNARY_OPS:
        return parse_unary(expr)
    if cmd_str in BOOL_OPS:
        return parse_boolop(expr)

    match cmd_str:
        case "defcontract":
            return parse_contract(expr)
        case 'defn':
            return parse_fn(expr)
        case 'return':
            return parse_return(expr)
        case 'quote':
            return parse_tuple(expr)
        case '.':
            return parse_attribute(expr)
        case 'setv':
            return parse_assignment(expr)
        case 'if':
            return parse_if(expr)
        case 'defvar':
            # return parse_declaration(expr[1], expr[2])
            return parse_declarations(expr)
        case _:
            return parse_call(expr)


def parse_node(node):
    match node:
        case models.Expression(node):
            return parse_expr(node)
        case models.Integer(node):
            value_node = vy_nodes.Int(value=int(node), node_id=next_nodeid(), ast_type='Int')
            return value_node
        case models.Float(node):
            raise Exception("Floating point not supported (yet)")
            # value_node = vy_nodes.Decimal(value=Decimal(float(node)), node_id=next_nodeid(), ast_type='Decimal')
            # return value_node
        case models.String(node):
            value_node = vy_nodes.Str(value=str(node), node_id=next_nodeid(), ast_type='Str')
            return value_node
        case models.Symbol(node) if str(node) in BUILTIN_FUNCS:
            return parse_builtin(node)
        case models.Symbol(node) if str(node) in NAME_CONSTS:
            return vy_nodes.NameConstant(value=py_ast.literal_eval(str(node)), id=next_nodeid(), ast_type='NameConstant')
        case models.Symbol(node) if str(node).startswith('0x'):
            return vy_nodes.Hex(id=next_nodeid(), ast_type='Hex', value=str(node))
        case models.Symbol(node) if "/" in str(node):
            target, attr = str(node).split('/')
            replacement_node = models.Expression((models.Symbol('.'), models.Symbol(target), models.Symbol(attr)))
            return parse_node(replacement_node)
        case models.Symbol(node) | models.Keyword(node):
            name_node = vy_nodes.Name(id=str(node), node_id=next_nodeid(), ast_type='Name')
            return name_node
        case models.Bytes(byt):
            bytes_node = vy_nodes.Bytes(node_id=next_nodeid(), ast_type='Byte', value=byt)
            return bytes_node
        case None:
            return None
        case _:
            raise Exception(f"No match for node {node}")

def parse_src(src: str):
    mod_node = vy_nodes.Module(body=[], name="", doc_string="", ast_type='Module', node_id=next_nodeid())

    vars = []
    fs = []
    for element in hy.read_many(src):
        ast = parse_node(element)

        if isinstance(ast, vy_nodes.Module):
            mod_node = ast
        elif isinstance(ast, vy_nodes.VariableDecl):
            vars.append(ast)
        elif isinstance(ast, vy_nodes.FunctionDef):
            fs.append(ast)
        elif isinstance(ast, list):
            for v in ast:
                vars.append(v)
        else:
            raise(f"Unrecognized top-level form {element} {ast}")

    for e in vars + fs:
        mod_node.add_to_body(e)


    return mod_node



from hy import models

def counter_gen():
    _counter = 0
    while True:
        yield _counter
        _counter += 1

def next_node_id_maker():
    counter = counter_gen()

    def next_num():
        return next(counter)

    return next_num

def pairwise(iterable):
    "s -> (s0, s1), (s2, s3), (s4, s5), ..."
    a = iter(iterable)
    return zip(a, a)

def has_return(tree):
    match tree:
        case models.Symbol(sym) if str(sym) == "return":
            return True
        case models.Sequence(seq):
            for el in seq:
                if has_return(el):
                    return True
        case _:
            return False
    return False


next_nodeid = next_node_id_maker()



from ast import Expression
import dasy
import vyper.ast.nodes as vy_nodes
from hy import models

from .utils import has_return, next_nodeid, pairwise


def parse_attribute(expr):
    match expr[1:]:
        case [obj, attr]:
            return vy_nodes.Attribute(ast_type='Attribute', node_id=next_nodeid(), attr=str(attr), value=dasy.parser.parse_node(obj))

def parse_call(expr):
    match expr:
        case (fn_name, *args):
            args_list = [dasy.parser.parse_node(arg) for arg in args]
            return vy_nodes.Call(func=dasy.parser.parse_node(fn_name), args=args_list, keywords=[], ast_type='Call', node_id=next_nodeid())


def parse_tuple(tuple_tree):
    match tuple_tree:
        case models.Symbol(q), elements if str(q) == 'quote':
            elts = [dasy.parser.parse_node(e) for e in elements]
            return vy_nodes.Tuple(elements=elts, node_id=next_nodeid(), ast_type='Tuple')
        case _:
            raise Exception("Invalid tuple declaration; requires quoted list ex: '(2 3 4)")

def parse_args_list(args_list) -> [vy_nodes.arg]:
    if len(args_list) == 0:
        return []
    results = []
    current_type = args_list[0]
    assert isinstance(current_type, models.Keyword) or isinstance(current_type, models.Expression)
    # get annotation and name
    for arg in args_list[1:]:
        # get annotation and name
        if isinstance(current_type, models.Keyword):
            annotation_node = vy_nodes.Name(id=str(current_type.name), parent=None, node_id=next_nodeid(), ast_type='Name')
        elif isinstance(current_type, models.Expression):
            annotation_node = dasy.parse.parse_node(current_type)
        results.append(vy_nodes.arg(arg=str(arg), parent=None, annotation=annotation_node, node_id=next_nodeid()))
    return results

def parse_fn(fn_tree):
    fn_node_id = next_nodeid()
    assert isinstance(fn_tree, models.Expression)
    assert fn_tree[0] == models.Symbol('defn')
    match fn_tree[1:]:
        case models.Symbol(sym_node), models.List(args_node), returns, decs, *body if isinstance(decs, models.Keyword) or isinstance(decs, models.List):
            assert isinstance(returns, models.Keyword) or isinstance(returns, models.Expression)
            rets = dasy.parse.parse_node(returns)
            name = str(sym_node)
            args_list = parse_args_list(args_node)
            args = vy_nodes.arguments(args=args_list, defaults=list(), node_id=next_nodeid(), ast_type='arguments')
            if isinstance(decs, models.Keyword):
                decorators = [vy_nodes.Name(id=str(decs.name), node_id=next_nodeid(), ast_type='Name')]
            elif isinstance(decs, models.List):
                decorators = [vy_nodes.Name(id=str(d.name), node_id=next_nodeid(), ast_type='Name') for d in decs]
            else:
                decorators = []
            fn_body = [dasy.parse.parse_node(body_node) for body_node in body[:-1]]
            if not has_return(body[-1]):
                value_node = dasy.parse.parse_node(body[-1])
                implicit_return_node = vy_nodes.Return(value=value_node, ast_type='Return', node_id=next_nodeid())
                fn_body.append(implicit_return_node)
            else:
                fn_body.append(dasy.parse.parse_node(body[-1]))
        case models.Symbol(sym_node), models.List(args_node), decs, *body:
            rets = None
            name = str(sym_node)
            args_list = parse_args_list(args_node)
            args = vy_nodes.arguments(args=args_list, defaults=list(), node_id=next_nodeid(), ast_type='arguments')
            if isinstance(decs, models.Keyword):
                decorators = [vy_nodes.Name(id=str(decs.name), node_id=next_nodeid(), ast_type='Name')]
            elif isinstance(decs, models.List):
                decorators = [vy_nodes.Name(id=str(d.name), node_id=next_nodeid(), ast_type='Name') for d in decs]
            else:
                decorators = []
            fn_body = [dasy.parse.parse_node(body_node) for body_node in body]
        case models.Symbol(sym_node), models.List(args_node), *body if str(sym_node) == "__init__":
            rets = None
            name = str(sym_node)
            args_list = parse_args_list(args_node)
            args = vy_nodes.arguments(args=args_list, defaults=list(), node_id=next_nodeid(), ast_type='arguments')
            decorators = [vy_nodes.Name(id="external", node_id=next_nodeid(), ast_type='Name')]
            fn_body = [dasy.parse.parse_node(body_node) for body_node in body]
        case _:
            raise Exception(f"Invalid fn form {fn_tree}")
    return vy_nodes.FunctionDef(args=args, returns=rets, decorator_list=decorators, pos=None, body=fn_body, name=name, node_id=fn_node_id, ast_type='FunctionDef')

def parse_declaration(var, typ):
    target = dasy.parse.parse_node(var)
    is_constant = False
    is_public = False
    is_immutable = False
    match typ:
        case [models.Symbol(e), typ_decl] if str(e) in ["public", "immutable", "constant"]:
            annotation = dasy.parse.parse_node(typ)
            match str(e):
                case "public":
                    is_public = True
                case "immutable":
                    is_immutable = True
                case "constant":
                    is_constant = True
        case models.Expression():
            annotation = dasy.parse.parse_node(typ)
        case models.Keyword():
            annotation = dasy.parse.parse_node(typ)
        case _:
            raise Exception(f"Invalid declaration type {typ}")
    return vy_nodes.VariableDecl(ast_type='VariableDecl', node_id=next_nodeid(), target=target, annotation=annotation, value=None, is_constant=is_constant, is_public=is_public, is_immutable=is_immutable)


def parse_declarations(expr):
    return [parse_declaration(var, typ) for var, typ in pairwise(expr[1:])]


def parse_contract(expr):
    mod_node = vy_nodes.Module(body=[], name=str(expr[1]), doc_string="", ast_type='Module', node_id=next_nodeid())
    expr_body = []
    match expr[1:]:
        case (name, vars, *body) if isinstance(vars, models.List):
            # # contract has state
            for var, typ in pairwise(vars):
            #     target = dasy.parse.parse_node(var)
            #     is_constant = False
            #     is_public = False
            #     is_immutable = False
            #     match typ:
            #         case [models.Symbol(e), typ_decl] if str(e) in ["public", "immutable", "constant"]:
            #             annotation = dasy.parse.parse_node(typ)
            #             match str(e):
            #                 case "public":
            #                     is_public = True
            #                 case "immutable":
            #                     is_immutable = True
            #                 case "constant":
            #                     is_constant = True
            #         case models.Keyword():
            #             annotation = dasy.parse.parse_node(typ)
            #         case _:
            #             raise Exception(f"Invalid declaration type {typ}")
            #     mod_node.add_to_body(vy_nodes.VariableDecl(ast_type='VariableDecl', node_id=next_nodeid(), target=target, annotation=annotation, value=None, is_constant=is_constant, is_public=is_public, is_immutable=is_immutable))
                mod_node.add_to_body(parse_declaration(var, typ))
            expr_body = expr[3:]
        case (name, *body):
            # no contract state
            expr_body = expr[2:]
        case _:
            # print(f"no match: {expr}")
            raise Exception(f"Invalid defcontract form: {expr}")
    for node in expr_body:
        mod_node.add_to_body(dasy.parse.parse_node(node))

    return mod_node



from hy import models
from dasy import parser
from vyper.ast.nodes import BinOp, Compare, UnaryOp, BoolOp
from .utils import next_nodeid

BIN_FUNCS = ['+', '-', '/', '*']
COMP_FUNCS = ['<', '<=', '>', '>=', '==', '!=']
UNARY_OPS = ['not']
BOOL_OPS = ['and', 'or']

def chain_comps(expr):
    new_node = models.Expression()
    new_expr = [models.Symbol("and")]
    for vals in zip(expr[1:], expr[2:]):
        new_expr.append(models.Expression((expr[0], vals[0], vals[1])))
    new_node += tuple(new_expr)
    return new_node


def parse_comparison(comp_tree):
    if len(comp_tree[1:]) > 2: # comparing more than 2 things; chain comps for (< 2 3 4 )
        return parser.parse_node(chain_comps(comp_tree))
    left = parser.parse_node(comp_tree[1])
    right = parser.parse_node(comp_tree[2])
    op = parser.parse_node(comp_tree[0])
    return Compare(left=left, ops=[op], comparators=[right], node_id=next_nodeid(), ast_type='Compare')

def parse_unary(expr):
    operand = parser.parse_node(expr[1])
    op = parser.parse_node(expr[0])
    return UnaryOp(operand=operand, op=op, node_id=next_nodeid(), ast_type="UnaryOp")

def parse_boolop(expr):
    op = parser.parse_node(expr[0])
    values = [parser.parse_node(e) for e in expr[1:]]
    return BoolOp(op=op, values=values, node_id=next_nodeid(), ast_type="BoolOp")

def chain_binops(expr):
    if len(expr) == 3:
        return expr
    else:
        new_node = models.Expression()
        tmp_expr = tuple([expr[0], *expr[2:]])
        tmp_node = models.Expression()
        tmp_node += tmp_expr
        subtree = chain_binops(tmp_node)
        new_node += tuple([expr[0], expr[1], subtree])
        return new_node


def parse_binop(binop_tree):
    match str(binop_tree[0]):
        case '+' | '-' | '*' | '/':
            if len(binop_tree) > 3:
                return parser.parse_node(chain_binops(binop_tree))
            left = parser.parse_node(binop_tree[1])
            right = parser.parse_node(binop_tree[2])
            op = parser.parse_node(binop_tree[0])
            return BinOp(left=left, right=right, op=op, node_id=next_nodeid(), ast_type='BinOp')



from dasy import compiler
import argparse
import sys

def main():
    parser = argparse.ArgumentParser(prog="dasy")
    parser.add_argument('filename', type=str, nargs='?', default="")

    src = ""

    args = parser.parse_args()

    if args.filename != "":
        with open(args.filename, 'r') as f:
            src = f.read()
    else:
        for line in sys.stdin:
            src += line

    data = compiler.compile(src)
    print("0x" + data.bytecode.hex())


if __name__ == "__main__":
    main()



import vyper.ast.nodes as vy_nodes
from .utils import next_nodeid

def parse_builtin(node):
    match str(node):
        case '+':
            op_node = vy_nodes.Add(node_id=next_nodeid(), ast_type='Add', _pretty="+", _description="addition")
            return op_node
        case '-':
            op_node = vy_nodes.Sub(node_id=next_nodeid(), ast_type='Sub', _pretty="-", _description="subtraction")
            return op_node
        case '*':
            op_node = vy_nodes.Mult(node_id=next_nodeid(), ast_type='Mult', _pretty="*", _description="multiplication")
            return op_node
        case '/':
            op_node = vy_nodes.Div(node_id=next_nodeid(), ast_type='Div', _pretty="/", _description="multiplication")
            return op_node
        case '<':
            op_node = vy_nodes.Lt(node_id=next_nodeid(), ast_type='Lt', _pretty="<", _description="less than")
            return op_node
        case '<=':
            op_node = vy_nodes.LtE(node_id=next_nodeid(), ast_type='LtE', _pretty="<=", _description="less than equal")
            return op_node
        case '>':
            op_node = vy_nodes.Gt(node_id=next_nodeid(), ast_type='Gt', _pretty="<", _description="greater than")
            return op_node
        case '>=':
            op_node = vy_nodes.GtE(node_id=next_nodeid(), ast_type='GtE', _pretty=">=", _description="greater than equal")
            return op_node
        case '==':
            op_node = vy_nodes.Eq(node_id=next_nodeid(), ast_type='Eq', _pretty="==", _description="equality")
            return op_node
        case '!=':
            op_node = vy_nodes.NotEq(node_id=next_nodeid(), ast_type='NotEq', _pretty="!=", _description="inequality")
            return op_node
        case 'not':
            op_node = vy_nodes.Not(node_id=next_nodeid(), ast_type='Not', _pretty="not", _description="negation")
            return op_node
        case 'and':
            op_node = vy_nodes.And(node_id=next_nodeid(), ast_type='And', _pretty="and", _description="boolean and")
            return op_node
        case 'or':
            op_node = vy_nodes.Or(node_id=next_nodeid(), ast_type='Or', _pretty="or", _description="boolean or")
            return op_node



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dasy', 'dasy.parser']

package_data = \
{'': ['*']}

install_requires = \
['argparse>=1.4.0,<2.0.0',
 'dasy-hy>=0.24.0,<0.25.0',
 'titanoboa>=0.1.5,<0.2.0',
 'vyper>=0.3.6,<0.4.0']

entry_points = \
{'console_scripts': ['dasy = dasy:main']}

setup_kwargs = {
    'name': 'dasy',
    'version': '0.1.11',
    'description': '',
    'long_description': None,
    'author': 'z80',
    'author_email': 'z80@ophy.xyz',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



from vyper.compiler.phases import CompilerData
from dasy.parser import parse_src

def compile(src: str) -> CompilerData:
    ast = parse_src(src)
    data = CompilerData("", ast.name, None, source_id=0, )
    data.vyper_module = ast
    return data
